@@ -89,6 +89,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 
+#include "ntp.h"
 #include "ntp_random.h"
 #include "ntp_stdlib.h"
 #include "ntp_assert.h"
@@ -107,15 +108,12 @@
 #endif	/* OPENSSL */
 #include <ssl_applink.c>
 
+#define _UC(str)	((char *)(intptr_t)(str))
 /*
  * Cryptodefines
  */
 #define	MD5KEYS		10	/* number of keys generated of each type */
 #define	MD5SIZE		20	/* maximum key size */
-#define DAY		((long)60*60*24) /* one day in seconds */
-#define	YEAR		((long)365) /* one year in days */
-#define MAXFILENAME	256	/* max file name length */
-#define MAXHOSTNAME	256	/* max host name length */
 #ifdef AUTOKEY
 #define	PLEN		512	/* default prime modulus size (bits) */
 #define	ILEN		256	/* default identity modulus size (bits) */
@@ -134,19 +132,19 @@
  * Prototypes
  */
 FILE	*fheader	(const char *, const char *, const char *);
-int	gen_md5		(char *);
+int	gen_md5		(const char *);
 void	followlink	(char *, size_t);
 #ifdef AUTOKEY
-EVP_PKEY *gen_rsa	(char *);
-EVP_PKEY *gen_dsa	(char *);
-EVP_PKEY *gen_iffkey	(char *);
-EVP_PKEY *gen_gqkey	(char *);
-EVP_PKEY *gen_mvkey	(char *, EVP_PKEY **);
+EVP_PKEY *gen_rsa	(const char *);
+EVP_PKEY *gen_dsa	(const char *);
+EVP_PKEY *gen_iffkey	(const char *);
+EVP_PKEY *gen_gqkey	(const char *);
+EVP_PKEY *gen_mvkey	(const char *, EVP_PKEY **);
 void	gen_mvserv	(char *, EVP_PKEY **);
-int	x509		(EVP_PKEY *, const EVP_MD *, char *, char *,
+int	x509		(EVP_PKEY *, const EVP_MD *, char *, const char *,
 			    char *);
 void	cb		(int, int, void *);
-EVP_PKEY *genkey	(char *, char *);
+EVP_PKEY *genkey	(const char *, const char *);
 EVP_PKEY *readkey	(char *, char *, u_int *, EVP_PKEY **);
 void	writekey	(char *, char *, u_int *, EVP_PKEY **);
 u_long	asn2ntp		(ASN1_TIME *);
@@ -157,13 +155,15 @@ u_long	asn2ntp		(ASN1_TIME *);
  */
 extern char *optarg;		/* command line argument */
 char	*progname;
-u_int	lifetime = YEAR;	/* certificate lifetime (days) */
+u_int	lifetime = DAYSPERYEAR;	/* certificate lifetime (days) */
 int	nkeys;			/* MV keys */
 time_t	epoch;			/* Unix epoch (seconds) since 1970 */
 u_int	fstamp;			/* NTP filestamp */
-char	*hostname = NULL;	/* host name */
+char	hostbuf[MAXHOSTNAME + 1];
+char	*hostname = NULL;	/* host, used in cert filenames */
 char	*groupname = NULL;	/* group name */
-char	*certname = NULL;	/* certificate subjetc/issuer name */
+char	certnamebuf[2 * sizeof(hostbuf)];
+char	*certname = NULL;	/* certificate subject/issuer name */
 char	*passwd1 = NULL;	/* input private key password */
 char	*passwd2 = NULL;	/* output private key password */
 char	filename[MAXFILENAME + 1]; /* file name */
@@ -171,29 +171,73 @@ char	filename[MAXFILENAME + 1]; /* file name */
 u_int	modulus = PLEN;		/* prime modulus size (bits) */
 u_int	modulus2 = ILEN;	/* identity modulus size (bits) */
 long	d0, d1, d2, d3;		/* callback counters */
+const EVP_CIPHER * cipher = NULL;
 #endif	/* AUTOKEY */
 
 #ifdef SYS_WINNT
 BOOL init_randfile();
 
 /*
- * Don't try to follow symbolic links.  Assumes link == file.
+ * Don't try to follow symbolic links on Windows.  Assume link == file.
  */
 int
-readlink(char *link, char *file, int len)
+readlink(
+	char *	link,
+	char *	file,
+	int	len
+	)
 {
 	return strlen(file);
 }
 
 /*
- * Don't try to create a symbolic link for now.
- * Just move the file to the name you need.
+ * Don't try to create symbolic links on Windows, that is supported on
+ * Vista and later only.  Instead, if CreateHardLink is available (XP
+ * and later), hardlink the linkname to the original filename.  On
+ * earlier systems, user must rename file to match expected link for
+ * ntpd to find it.  To allow building a ntp-keygen.exe which loads on
+ * Windows pre-XP, runtime link to CreateHardLinkA().
  */
 int
-symlink(char *filename, char *linkname) {
-	DeleteFile(linkname);
-	MoveFile(filename, linkname);
-	return (0);
+symlink(
+	char *	filename,
+	char*	linkname
+	)
+{
+	typedef BOOL (WINAPI *PCREATEHARDLINKA)(
+		__in LPCSTR	lpFileName,
+		__in LPCSTR	lpExistingFileName,
+		__reserved LPSECURITY_ATTRIBUTES lpSA
+		);
+	static PCREATEHARDLINKA pCreateHardLinkA;
+	static int		tried;
+	HMODULE			hDll;
+	FARPROC			pfn;
+	int			link_created;
+	int			saved_errno;
+
+	if (!tried) {
+		tried = TRUE;
+		hDll = LoadLibrary("kernel32");
+		pfn = GetProcAddress(hDll, "CreateHardLinkA");
+		pCreateHardLinkA = (PCREATEHARDLINKA)pfn;
+	}
+
+	if (NULL == pCreateHardLinkA) {
+		errno = ENOSYS;
+		return -1;
+	}
+
+	link_created = (*pCreateHardLinkA)(linkname, filename, NULL);
+	
+	if (link_created)
+		return 0;
+
+	saved_errno = GetLastError();	/* yes we play loose */
+	mfprintf(stderr, "Create hard link %s to %s failed: %m\n",
+		 linkname, filename);
+	errno = saved_errno;
+	return -1;
 }
 
 void
@@ -223,9 +267,13 @@ followlink(
 {
 	int len;
 
+	REQUIRE(bufsiz > 0);
+
 	len = readlink(fname, fname, (int)bufsiz);
-	if (len < 0 || bufsiz < 1)
+	if (len < 0 ) {
+		fname[0] = '\0';
 		return;
+	}
 	if (len > (int)bufsiz - 1)
 		len = (int)bufsiz - 1;
 	fname[len] = '\0';
@@ -243,6 +291,7 @@ main(
 {
 	struct timeval tv;	/* initialization vector */
 	int	md5key = 0;	/* generate MD5 keys */
+	int	optct;		/* option count */
 #ifdef AUTOKEY
 	X509	*cert = NULL;	/* X509 certificate */
 	X509_EXTENSION *ext;	/* X509v3 extension */
@@ -261,16 +310,18 @@ main(
 	EVP_PKEY *pkey = NULL;	/* temp key */
 	const EVP_MD *ectx;	/* EVP digest */
 	char	pathbuf[MAXFILENAME + 1];
-	char	str[MAXFILENAME + 1];
 	const char *scheme = NULL; /* digest/signature scheme */
-	char	*exten = NULL;	/* private extension */
+	const char *ciphername = NULL; /* to encrypt priv. key */
+	const char *exten = NULL;	/* private extension */
 	char	*grpkey = NULL;	/* identity extension */
 	int	nid;		/* X509 digest/signature scheme */
 	FILE	*fstr = NULL;	/* file handle */
 	char	groupbuf[MAXHOSTNAME + 1];
-#define iffsw   HAVE_OPT(ID_KEY)
+	u_int	temp;
+	BIO *	bp;
+	int	i, cnt;
+	char *	ptr;
 #endif	/* AUTOKEY */
-	char	hostbuf[MAXHOSTNAME + 1];
 
 	progname = argv[0];
 
@@ -286,22 +337,26 @@ main(
 	ssl_check_version();
 #endif	/* OPENSSL */
 
+	ntp_crypto_srandom();
+
 	/*
 	 * Process options, initialize host name and timestamp.
+	 * gethostname() won't null-terminate if hostname is exactly the
+	 * length provided for the buffer.
 	 */
-	gethostname(hostbuf, MAXHOSTNAME);
-	hostname = groupname = certname = passwd1 = hostbuf;
+	gethostname(hostbuf, sizeof(hostbuf) - 1);
+	hostbuf[COUNTOF(hostbuf) - 1] = '\0';
+	hostname = hostbuf;
+	groupname = hostbuf;
+	passwd1 = hostbuf;
 	passwd2 = NULL;
-	gettimeofday(&tv, 0);
+	GETTIMEOFDAY(&tv, NULL);
 	epoch = tv.tv_sec;
 	fstamp = (u_int)(epoch + JAN_1970);
 
-	{
-		int optct = ntpOptionProcess(&ntp_keygenOptions,
-					     argc, argv);
-		argc -= optct;
-		argv += optct;
-	}
+	optct = ntpOptionProcess(&ntp_keygenOptions, argc, argv);
+	argc -= optct;
+	argv += optct;
 
 #ifdef OPENSSL
 	if (SSLeay() == SSLEAY_VERSION_NUMBER)
@@ -312,16 +367,16 @@ main(
 			OPENSSL_VERSION_TEXT, SSLeay_version(SSLEAY_VERSION));
 #endif /* OPENSSL */
 
-	debug = DESC(DEBUG_LEVEL).optOccCt;
+	debug = OPT_VALUE_SET_DEBUG_LEVEL;
+
 	if (HAVE_OPT( MD5KEY ))
 		md5key++;
-
 #ifdef AUTOKEY
-	if (HAVE_OPT( PVT_PASSWD ))
-		passwd1 = estrdup(OPT_ARG( PVT_PASSWD ));
+	if (HAVE_OPT( PASSWORD ))
+		passwd1 = estrdup(OPT_ARG( PASSWORD ));
 
-	if (HAVE_OPT( GET_PVT_PASSWD ))
-		passwd2 = estrdup(OPT_ARG( GET_PVT_PASSWD ));
+	if (HAVE_OPT( EXPORT_PASSWD ))
+		passwd2 = estrdup(OPT_ARG( EXPORT_PASSWD ));
 
 	if (HAVE_OPT( HOST_KEY ))
 		hostkey++;
@@ -343,24 +398,24 @@ main(
 		mvpar++;
 		nkeys = OPT_VALUE_MV_KEYS;
 	}
+
+	if (HAVE_OPT( IMBITS ))
+		modulus2 = OPT_VALUE_IMBITS;
+
 	if (HAVE_OPT( MODULUS ))
 		modulus = OPT_VALUE_MODULUS;
 
 	if (HAVE_OPT( CERTIFICATE ))
 		scheme = OPT_ARG( CERTIFICATE );
 
-	if (HAVE_OPT( SUBJECT_NAME )) {
-		if (*OPT_ARG(SUBJECT_NAME) != '@') {
-			certname = estrdup(OPT_ARG(SUBJECT_NAME));
-		} else {
-			strlcpy(str, certname, sizeof(str));
-			strlcat(str, OPT_ARG(SUBJECT_NAME), sizeof(str));
-			certname = estrdup(str);
-		}
-	}
+	if (HAVE_OPT( CIPHER ))
+		ciphername = OPT_ARG( CIPHER );
 
-	if (HAVE_OPT( ISSUER_NAME ))
-		groupname = estrdup(OPT_ARG( ISSUER_NAME ));
+	if (HAVE_OPT( SUBJECT_NAME ))
+		hostname = estrdup(OPT_ARG( SUBJECT_NAME ));
+
+	if (HAVE_OPT( IDENT ))
+		groupname = estrdup(OPT_ARG( IDENT ));
 
 	if (HAVE_OPT( LIFETIME ))
 		lifetime = OPT_VALUE_LIFETIME;
@@ -371,14 +426,43 @@ main(
 	if (HAVE_OPT( TRUSTED_CERT ))
 		exten = EXT_KEY_TRUST;
 
+	/*
+	 * Remove the group name from the hostname variable used
+	 * in host and sign certificate file names.
+	 */
+	if (hostname != hostbuf)
+		ptr = strchr(hostname, '@');
+	else
+		ptr = NULL;
+	if (ptr != NULL) {
+		*ptr = '\0';
+		groupname = estrdup(ptr + 1);
+		/* -s @group is equivalent to -i group, host unch. */
+		if (ptr == hostname)
+			hostname = hostbuf;
+	}
+
+	/*
+	 * Derive host certificate issuer/subject names from host name
+	 * and optional group.  If no groupname is provided, the issuer
+	 * and subject is the hostname with no '@group', and the
+	 * groupname variable is pointed to hostname for use in IFF, GQ,
+	 * and MV parameters file names.
+	 */
+	if (groupname == hostbuf) {
+		certname = hostname;
+	} else {
+		snprintf(certnamebuf, sizeof(certnamebuf), "%s@%s",
+			 hostname, groupname);
+		certname = certnamebuf;
+	}
+
 	/*
 	 * Seed random number generator and grow weeds.
 	 */
 	ERR_load_crypto_strings();
 	OpenSSL_add_all_algorithms();
 	if (!RAND_status()) {
-		u_int	temp;
-
 		if (RAND_file_name(pathbuf, sizeof(pathbuf)) == NULL) {
 			fprintf(stderr, "RAND_file_name %s\n",
 			    ERR_error_string(ERR_get_error(), NULL));
@@ -395,7 +479,18 @@ main(
 		    "Random seed file %s %u bytes\n", pathbuf, temp);
 		RAND_add(&epoch, sizeof(epoch), 4.0);
 	}
+#endif	/* AUTOKEY */
 
+	/*
+	 * Create new unencrypted MD5 keys file if requested. If this
+	 * option is selected, ignore all other options.
+	 */
+	if (md5key) {
+		gen_md5("md5");
+		exit (0);
+	}
+
+#ifdef AUTOKEY
 	/*
 	 * Load previous certificate if available.
 	 */
@@ -426,10 +521,6 @@ main(
 		 * whether this is a trusted or private certificate.
 		 */
 		if (exten == NULL) {
-			BIO	*bp;
-			int	i, cnt;
-			char	*ptr;
-
 			ptr = strstr(groupbuf, "CN=");
 			cnt = X509_get_ext_count(cert);
 			for (i = 0; i < cnt; i++) {
@@ -447,28 +538,23 @@ main(
 					else if (strcmp(pathbuf,
 					    "Private") == 0)
 						exten = EXT_KEY_PRIVATE;
-					if (certname == NULL)
-						certname = ptr + 3;
+					certname = estrdup(ptr + 3);
 				}
 			}
 		}
 	}
 	if (scheme == NULL)
 		scheme = "RSA-MD5";
+	if (ciphername == NULL)
+		ciphername = "des-ede3-cbc";
+	cipher = EVP_get_cipherbyname(ciphername);
+	if (cipher == NULL) {
+		fprintf(stderr, "Unknown cipher %s\n", ciphername);
+		exit(-1);
+	}
 	fprintf(stderr, "Using host %s group %s\n", hostname,
 	    groupname);
-#endif	/* AUTOKEY */
 
-	/*
-	 * Create new unencrypted MD5 keys file if requested. If this
-	 * option is selected, ignore all other options.
-	 */
-	if (md5key) {
-		gen_md5("md5");
-		exit (0);
-	}
-
-#ifdef AUTOKEY
 	/*
 	 * Create a new encrypted RSA host key file if requested;
 	 * otherwise, look for an existing host key file. If not found,
@@ -490,7 +576,7 @@ main(
 	}
 	if (pkey_host == NULL) {
 		fprintf(stderr, "Generating host key fails\n");
-		exit (-1);
+		exit(-1);
 	}
 
 	/*
@@ -553,9 +639,9 @@ main(
 		BN_copy(rsa->q, BN_value_one());
 		pkey = EVP_PKEY_new();
 		EVP_PKEY_assign_RSA(pkey, rsa);
-		PEM_write_PrivateKey(stdout, pkey, NULL, NULL, 0, NULL,
-		    NULL);
-		fclose(stdout);
+		PEM_write_PKCS8PrivateKey(stdout, pkey, NULL, NULL, 0,
+		    NULL, NULL);
+		fflush(stdout);
 		if (debug)
 			RSA_print_fp(stderr, rsa, 0);
 	}
@@ -575,9 +661,9 @@ main(
 		rsa = pkey_gqkey->pkey.rsa;
 		pkey = EVP_PKEY_new();
 		EVP_PKEY_assign_RSA(pkey, rsa);
-		PEM_write_PrivateKey(stdout, pkey,
-		    EVP_des_cbc(), NULL, 0, NULL, passwd2);
-		fclose(stdout);
+		PEM_write_PKCS8PrivateKey(stdout, pkey, cipher, NULL, 0,
+		    NULL, passwd2);
+		fflush(stdout);
 		if (debug)
 			RSA_print_fp(stderr, rsa, 0);
 	}
@@ -617,9 +703,9 @@ main(
 		BN_copy(dsa->priv_key, BN_value_one());
 		pkey = EVP_PKEY_new();
 		EVP_PKEY_assign_DSA(pkey, dsa);
-		PEM_write_PrivateKey(stdout, pkey, NULL, NULL, 0, NULL,
-		    NULL);
-		fclose(stdout);
+		PEM_write_PKCS8PrivateKey(stdout, pkey, NULL, NULL, 0,
+		    NULL, NULL);
+		fflush(stdout);
 		if (debug)
 			DSA_print_fp(stderr, dsa, 0);
 	}
@@ -639,9 +725,9 @@ main(
 		dsa = pkey_iffkey->pkey.dsa;
 		pkey = EVP_PKEY_new();
 		EVP_PKEY_assign_DSA(pkey, dsa);
-		PEM_write_PrivateKey(stdout, pkey, EVP_des_cbc(), NULL,
-		    0, NULL, passwd2);
-		fclose(stdout);
+		PEM_write_PKCS8PrivateKey(stdout, pkey, cipher, NULL, 0,
+		    NULL, passwd2);
+		fflush(stdout);
 		if (debug)
 			DSA_print_fp(stderr, dsa, 0);
 	}
@@ -677,9 +763,9 @@ main(
 		fprintf(stdout, "# %s\n# %s\n", filename,
 		    ctime(&epoch));
 		pkey = pkey_mvpar[2];
-		PEM_write_PrivateKey(stdout, pkey, NULL, NULL, 0, NULL,
-		    NULL);
-		fclose(stdout);
+		PEM_write_PKCS8PrivateKey(stdout, pkey, NULL, NULL, 0,
+		    NULL, NULL);
+		fflush(stdout);
 		if (debug)
 			DSA_print_fp(stderr, pkey->pkey.dsa, 0);
 	}
@@ -695,9 +781,9 @@ main(
 		fprintf(stdout, "# %s\n# %s\n", filename,
 		    ctime(&epoch));
 		pkey = pkey_mvpar[1];
-		PEM_write_PrivateKey(stdout, pkey, EVP_des_cbc(), NULL,
-		    0, NULL, passwd2);
-		fclose(stdout);
+		PEM_write_PKCS8PrivateKey(stdout, pkey, cipher, NULL, 0,
+		    NULL, passwd2);
+		fflush(stdout);
 		if (debug)
 			DSA_print_fp(stderr, pkey->pkey.dsa, 0);
 	}
@@ -715,7 +801,7 @@ main(
 	}
 	x509(pkey_sign, ectx, grpkey, exten, certname);
 #endif	/* AUTOKEY */
-	exit (0);
+	exit(0);
 }
 
 
@@ -726,7 +812,7 @@ main(
  */
 int
 gen_md5(
-	char	*id		/* file name id */
+	const char *id		/* file name id */
 	)
 {
 	u_char	md5key[MD5SIZE + 1];	/* MD5 key */
@@ -739,13 +825,19 @@ gen_md5(
 #endif	/* OPENSSL */
 
 	str = fheader("MD5key", id, groupname);
-	ntp_srandom((u_long)epoch);
 	for (i = 1; i <= MD5KEYS; i++) {
 		for (j = 0; j < MD5SIZE; j++) {
 			int temp;
 
 			while (1) {
-				temp = ntp_random() & 0xff;
+				int rc;
+
+				rc = ntp_crypto_random_buf(&temp, 1);
+				if (-1 == rc) {
+					fprintf(stderr, "ntp_crypto_random_buf() failed.\n");
+					exit (-1);
+				}
+				temp &= 0xff;
 				if (temp == '#')
 					continue;
 
@@ -867,7 +959,7 @@ readkey(
  */
 EVP_PKEY *			/* public/private key pair */
 gen_rsa(
-	char	*id		/* file name id */
+	const char *id		/* file name id */
 	)
 {
 	EVP_PKEY *pkey;		/* private key */
@@ -875,7 +967,7 @@ gen_rsa(
 	FILE	*str;
 
 	fprintf(stderr, "Generating RSA keys (%d bits)...\n", modulus);
-	rsa = RSA_generate_key(modulus, 3, cb, "RSA");
+	rsa = RSA_generate_key(modulus, 65537, cb, _UC("RSA"));
 	fprintf(stderr, "\n");
 	if (rsa == NULL) {
 		fprintf(stderr, "RSA generate keys fails\n%s\n",
@@ -906,7 +998,7 @@ gen_rsa(
 		str = fheader("RSAhost", id, hostname);
 	pkey = EVP_PKEY_new();
 	EVP_PKEY_assign_RSA(pkey, rsa);
-	PEM_write_PrivateKey(str, pkey, EVP_des_cbc(), NULL, 0, NULL,
+	PEM_write_PKCS8PrivateKey(str, pkey, cipher, NULL, 0, NULL,
 	    passwd1);
 	fclose(str);
 	if (debug)
@@ -920,7 +1012,7 @@ gen_rsa(
  */
 EVP_PKEY *			/* public/private key pair */
 gen_dsa(
-	char	*id		/* file name id */
+	const char *id		/* file name id */
 	)
 {
 	EVP_PKEY *pkey;		/* private key */
@@ -935,7 +1027,7 @@ gen_dsa(
 	    "Generating DSA parameters (%d bits)...\n", modulus);
 	RAND_bytes(seed, sizeof(seed));
 	dsa = DSA_generate_parameters(modulus, seed, sizeof(seed), NULL,
-	    NULL, cb, "DSA");
+	    NULL, cb, _UC("DSA"));
 	fprintf(stderr, "\n");
 	if (dsa == NULL) {
 		fprintf(stderr, "DSA generate parameters fails\n%s\n",
@@ -961,7 +1053,7 @@ gen_dsa(
 	str = fheader("DSAsign", id, hostname);
 	pkey = EVP_PKEY_new();
 	EVP_PKEY_assign_DSA(pkey, dsa);
-	PEM_write_PrivateKey(str, pkey, EVP_des_cbc(), NULL, 0, NULL,
+	PEM_write_PKCS8PrivateKey(str, pkey, cipher, NULL, 0, NULL,
 	    passwd1);
 	fclose(str);
 	if (debug)
@@ -1022,7 +1114,7 @@ gen_dsa(
  */
 EVP_PKEY *			/* DSA cuckoo nest */
 gen_iffkey(
-	char	*id		/* file name id */
+	const char *id		/* file name id */
 	)
 {
 	EVP_PKEY *pkey;		/* private key */
@@ -1040,7 +1132,7 @@ gen_iffkey(
 	    modulus2);
 	RAND_bytes(seed, sizeof(seed));
 	dsa = DSA_generate_parameters(modulus2, seed, sizeof(seed), NULL,
-	    NULL, cb, "IFF");
+	    NULL, cb, _UC("IFF"));
 	fprintf(stderr, "\n");
 	if (dsa == NULL) {
 		fprintf(stderr, "DSA generate parameters fails\n%s\n",
@@ -1128,7 +1220,7 @@ gen_iffkey(
 	str = fheader("IFFkey", id, groupname);
 	pkey = EVP_PKEY_new();
 	EVP_PKEY_assign_DSA(pkey, dsa);
-	PEM_write_PrivateKey(str, pkey, EVP_des_cbc(), NULL, 0, NULL,
+	PEM_write_PKCS8PrivateKey(str, pkey, cipher, NULL, 0, NULL,
 	    passwd1);
 	fclose(str);
 	if (debug)
@@ -1200,7 +1292,7 @@ gen_iffkey(
  */
 EVP_PKEY *			/* RSA cuckoo nest */
 gen_gqkey(
-	char	*id		/* file name id */
+	const char *id		/* file name id */
 	)
 {
 	EVP_PKEY *pkey;		/* private key */
@@ -1216,7 +1308,7 @@ gen_gqkey(
 	fprintf(stderr,
 	    "Generating GQ parameters (%d bits)...\n",
 	     modulus2);
-	rsa = RSA_generate_key(modulus2, 3, cb, "GQ");
+	rsa = RSA_generate_key(modulus2, 65537, cb, _UC("GQ"));
 	fprintf(stderr, "\n");
 	if (rsa == NULL) {
 		fprintf(stderr, "RSA generate keys fails\n%s\n",
@@ -1325,7 +1417,7 @@ gen_gqkey(
 	str = fheader("GQkey", id, groupname);
 	pkey = EVP_PKEY_new();
 	EVP_PKEY_assign_RSA(pkey, rsa);
-	PEM_write_PrivateKey(str, pkey, EVP_des_cbc(), NULL, 0, NULL,
+	PEM_write_PKCS8PrivateKey(str, pkey, cipher, NULL, 0, NULL,
 	    passwd1);
 	fclose(str);
 	if (debug)
@@ -1409,7 +1501,7 @@ gen_gqkey(
  */
 EVP_PKEY *			/* DSA cuckoo nest */
 gen_mvkey(
-	char	*id,		/* file name id */
+	const char *id,		/* file name id */
 	EVP_PKEY **evpars	/* parameter list pointer */
 	)
 {
@@ -1727,7 +1819,7 @@ gen_mvkey(
 	BN_copy(dsa->pub_key, b);
 	pkey = EVP_PKEY_new();
 	EVP_PKEY_assign_DSA(pkey, dsa);
-	PEM_write_PrivateKey(str, pkey, EVP_des_cbc(), NULL, 0, NULL,
+	PEM_write_PKCS8PrivateKey(str, pkey, cipher, NULL, 0, NULL,
 	    passwd1);
 	evpars[i++] = pkey;
 	if (debug)
@@ -1753,7 +1845,7 @@ gen_mvkey(
 	dsa2->pub_key = BN_dup(ghat);
 	pkey1 = EVP_PKEY_new();
 	EVP_PKEY_assign_DSA(pkey1, dsa2);
-	PEM_write_PrivateKey(str, pkey1, EVP_des_cbc(), NULL, 0, NULL,
+	PEM_write_PKCS8PrivateKey(str, pkey1, cipher, NULL, 0, NULL,
 	    passwd1);
 	evpars[i++] = pkey1;
 	if (debug)
@@ -1778,7 +1870,7 @@ gen_mvkey(
 		sdsa->pub_key = BN_dup(xhat[j]);
 		pkey1 = EVP_PKEY_new();
 		EVP_PKEY_set1_DSA(pkey1, sdsa);
-		PEM_write_PrivateKey(str, pkey1, EVP_des_cbc(), NULL, 0,
+		PEM_write_PKCS8PrivateKey(str, pkey1, cipher, NULL, 0,
 		    NULL, passwd1);
 		evpars[i++] = pkey1;
 		if (debug)
@@ -1831,10 +1923,10 @@ gen_mvkey(
  */
 int
 x509	(
-	EVP_PKEY *pkey,		/* generic signature algorithm */
-	const EVP_MD *md,	/* generic digest algorithm */
+	EVP_PKEY *pkey,		/* signing key */
+	const EVP_MD *md,	/* signature/digest scheme */
 	char	*gqpub,		/* identity extension (hex string) */
-	char	*exten,		/* private cert extension */
+	const char *exten,	/* private cert extension */
 	char	*name		/* subject/issuer name */
 	)
 {
@@ -1862,15 +1954,15 @@ x509	(
 	X509_set_serialNumber(cert, serial);
 	ASN1_INTEGER_free(serial);
 	X509_time_adj(X509_get_notBefore(cert), 0L, &epoch);
-	X509_time_adj(X509_get_notAfter(cert), lifetime * DAY, &epoch);
+	X509_time_adj(X509_get_notAfter(cert), lifetime * SECSPERDAY, &epoch);
 	subj = X509_get_subject_name(cert);
 	X509_NAME_add_entry_by_txt(subj, "commonName", MBSTRING_ASC,
-	    (unsigned char *) name, strlen(name), -1, 0);
+	    (u_char *)name, strlen(name), -1, 0);
 	subj = X509_get_issuer_name(cert);
 	X509_NAME_add_entry_by_txt(subj, "commonName", MBSTRING_ASC,
-	    (unsigned char *) name, strlen(name), -1, 0);
+	    (u_char *)name, strlen(name), -1, 0);
 	if (!X509_set_pubkey(cert, pkey)) {
-		fprintf(stderr, "Assign key fails\n%s\n",
+		fprintf(stderr, "Assign certificate signing key fails\n%s\n",
 		    ERR_error_string(ERR_get_error(), NULL));
 		X509_free(cert);
 		return (0);
@@ -1888,7 +1980,7 @@ x509	(
 	fprintf(stderr, "%s: %s\n", LN_basic_constraints,
 	    BASIC_CONSTRAINTS);
 	ex = X509V3_EXT_conf_nid(NULL, NULL, NID_basic_constraints,
-	    BASIC_CONSTRAINTS);
+	    _UC(BASIC_CONSTRAINTS));
 	if (!X509_add_ext(cert, ex, -1)) {
 		fprintf(stderr, "Add extension field fails\n%s\n",
 		    ERR_error_string(ERR_get_error(), NULL));
@@ -1901,7 +1993,7 @@ x509	(
 	 * be used for.
 	 */
 	fprintf(stderr, "%s: %s\n", LN_key_usage, KEY_USAGE);
-	ex = X509V3_EXT_conf_nid(NULL, NULL, NID_key_usage, KEY_USAGE);
+	ex = X509V3_EXT_conf_nid(NULL, NULL, NID_key_usage, _UC(KEY_USAGE));
 	if (!X509_add_ext(cert, ex, -1)) {
 		fprintf(stderr, "Add extension field fails\n%s\n",
 		    ERR_error_string(ERR_get_error(), NULL));
@@ -1936,7 +2028,7 @@ x509	(
 	if (exten != NULL) {
 		fprintf(stderr, "%s: %s\n", LN_ext_key_usage, exten);
 		ex = X509V3_EXT_conf_nid(NULL, NULL,
-		    NID_ext_key_usage, exten);
+		    NID_ext_key_usage, _UC(exten));
 		if (!X509_add_ext(cert, ex, -1)) {
 			fprintf(stderr,
 			    "Add extension field fails\n%s\n",
@@ -2047,8 +2139,8 @@ cb	(
  */
 EVP_PKEY *			/* public/private key pair */
 genkey(
-	char	*type,		/* key type (RSA or DSA) */
-	char	*id		/* file name id */
+	const char *type,	/* key type (RSA or DSA) */
+	const char *id		/* file name id */
 	)
 {
 	if (type == NULL)
@@ -2087,7 +2179,7 @@ fheader	(
 	}
 	snprintf(linkname, sizeof(linkname), "ntpkey_%s_%s", ulink,
 	    hostname);
-	remove(linkname);
+	(void)remove(linkname);		/* The symlink() line below matters */
 	temp = symlink(filename, linkname);
 	if (temp < 0)
 		perror(file);
